use core::fmt::Write;

use embedded_graphics::{
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::Rgb565,
    prelude::{Point, RgbColor, Size},
    primitives::{Primitive, PrimitiveStyleBuilder, Rectangle},
    text::{Baseline, Text},
    Drawable,
};
use heapless::String;

use crate::{app::WIDTH, storage::Storage};

use super::scroll_tracker::{ScrollAction, ScrollTracker};

const STATUS_BAR_HEIGHT: i32 = 40;
const GROUP_HEIGHT: i32 = 40;
const BORDER_WIDTH: u32 = 4;
const BACKGROUND_COLOR: Rgb565 = Rgb565::new(31, 41, 0); // #FFA500
const BORDER_COLOR: Rgb565 = Rgb565::new(31, 26, 0); // #FF6600

pub struct ChatList {
    tracker: ScrollTracker,
    selected: usize,
    len: usize,
}

impl ChatList {
    pub fn new(storage: &Storage, selected: usize) -> Self {
        let len = storage.message_addressees().count();

        Self {
            tracker: ScrollTracker::new(),
            selected,
            len,
        }
    }

    pub fn selected(&self) -> usize {
        self.selected
    }

    pub fn render<
        E,
        DT: embedded_graphics::prelude::DrawTarget<
            Color = embedded_graphics::pixelcolor::Rgb565,
            Error = E,
        >,
    >(
        &mut self,
        dt: &mut DT,
        storage: &Storage,
    ) -> Result<(), E> {
        let unselected_rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(BORDER_COLOR)
            .stroke_width(BORDER_WIDTH)
            .fill_color(BACKGROUND_COLOR)
            .build();

        let selected_rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(BORDER_COLOR)
            .stroke_width(BORDER_WIDTH)
            .fill_color(Rgb565::WHITE)
            .build();

        for (idx, (callsign, ssid)) in storage.message_addressees().enumerate() {
            let i = i32::try_from(idx).unwrap();

            let x = 0;
            let y = i * (GROUP_HEIGHT + i32::try_from(BORDER_WIDTH).unwrap()) + STATUS_BAR_HEIGHT;

            // Draw outer rectangle
            let rect_style = if idx == self.selected {
                selected_rect_style
            } else {
                unselected_rect_style
            };

            Rectangle::new(
                Point::new(0, y),
                Size::new(
                    u32::try_from(WIDTH).unwrap(),
                    u32::try_from(GROUP_HEIGHT).unwrap(),
                ),
            )
            .into_styled(rect_style)
            .draw(dt)?;

            // Draw text
            let text_x = x + i32::try_from(BORDER_WIDTH).unwrap();
            let text_y = y + i32::try_from(BORDER_WIDTH).unwrap();

            let text_style = MonoTextStyleBuilder::new()
                .font(&FONT_10X20)
                .text_color(Rgb565::BLACK)
                .build();

            let name = storage
                .contacts()
                .find(|c| c.callsign == callsign && c.ssid == ssid)
                .map(|c| c.name);

            let mut text: String<25> = String::new();
            match name {
                Some(x) => {
                    text = String::try_from(x.as_str()).unwrap();
                }
                None => {
                    write!(&mut text, "{}-{}", callsign, ssid).unwrap();
                }
            };
            Text::with_baseline(&text, Point::new(text_x, text_y), text_style, Baseline::Top)
                .draw(dt)?;
        }

        Ok(())
    }

    pub fn key_push(&mut self, _: crate::keyboard::KeyCode) {}

    pub fn touchpad_scroll(&mut self, x: i8, y: i8) {
        let (_, scroll_y) = self.tracker.scroll(x, y);
        match scroll_y {
            ScrollAction::Previous => {
                if self.selected > 0 {
                    self.selected -= 1;
                }
            }
            ScrollAction::Next => {
                if self.selected < self.len - 1 {
                    self.selected += 1;
                }
            }
            ScrollAction::None => {}
        }
    }
}
