use embedded_graphics::{
    geometry::Point,
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::{Rgb565, RgbColor, WebColors},
    text::{Baseline, Text},
    Drawable,
};

use crate::storage::{Settings, Storage, MAX_CONTACT_CALLSIGN_LENGTH};

use super::{
    scroll_tracker::{ScrollAction, ScrollTracker},
    slider::Slider,
    textbox::{self, TextBox},
    Element,
};

const LABEL_X: i32 = 10;
const TEXT_X: i32 = 140;

pub struct SettingsView {
    callsign: TextBox<MAX_CONTACT_CALLSIGN_LENGTH>,
    ssid: TextBox<3>,
    icon: TextBox<5>,
    freq: TextBox<9>,
    brightness: Slider,

    tracker: ScrollTracker,
    selected: u8,
}

impl SettingsView {
    pub fn new(storage: &Storage) -> Self {
        let settings = storage.settings();

        let mut callsign = TextBox::new(
            TEXT_X,
            (40 - textbox::BORDER_WIDTH).try_into().unwrap(),
            u32::try_from(MAX_CONTACT_CALLSIGN_LENGTH).unwrap() + 1,
            1,
            1,
            false,
        );
        callsign.set_text(settings.callsign()).unwrap(); // safe to unwrap - same length so it'll definitely fit

        let mut ssid = TextBox::new(
            TEXT_X,
            (80 - textbox::BORDER_WIDTH).try_into().unwrap(),
            4,
            1,
            1,
            true,
        );
        ssid.selected = false;
        ssid.set_text(&settings.ssid_string()).unwrap();

        let mut icon = TextBox::new(
            TEXT_X,
            (120 - textbox::BORDER_WIDTH).try_into().unwrap(),
            6,
            1,
            1,
            true,
        );
        icon.selected = false;
        icon.set_text(&settings.icon_string()).unwrap();

        let mut freq = TextBox::new(
            TEXT_X,
            (160 - textbox::BORDER_WIDTH).try_into().unwrap(),
            10,
            1,
            1,
            true,
        );
        freq.selected = false;
        freq.set_text(&settings.freq_string()).unwrap();

        let brightness = Slider::new(
            0,
            100,
            TEXT_X,
            (200 - textbox::BORDER_WIDTH).try_into().unwrap(),
        );

        Self {
            callsign,
            ssid,
            icon,
            freq,
            brightness,

            tracker: ScrollTracker::new(),
            selected: 0,
        }
    }

    pub fn exit(&self, storage: &mut Storage) -> bool {
        let (Some(ssid), Some(icon), Some(freq)) = (self.ssid(), self.icon(), self.freq()) else {
            return false;
        };

        let settings = storage.settings_mut();
        settings.set_callsign(self.callsign.get_text()).unwrap(); // safe to unwrap because we know the length fits
        settings.set_ssid(ssid);
        settings.set_icon(icon);
        settings.set_frequency(freq).unwrap(); // safe to unwrap because we already verified it's in range

        // settings.set_brightness(self.brightness); TODO
        storage.save_settings();

        true
    }

    fn ssid(&self) -> Option<u8> {
        self.ssid.get_text().parse().ok()
    }

    fn icon(&self) -> Option<u16> {
        self.icon.get_text().parse().ok()
    }

    fn freq(&self) -> Option<u32> {
        let Ok(freq): Result<u32, _> = self.freq.get_text().parse() else {
            return None;
        };

        if Settings::validate_freq(freq) {
            Some(freq)
        } else {
            None
        }
    }

    pub fn render<
        E,
        DT: embedded_graphics::prelude::DrawTarget<
            Color = embedded_graphics::pixelcolor::Rgb565,
            Error = E,
        >,
    >(
        &mut self,
        dt: &mut DT,
    ) -> Result<(), E> {
        let label_style = MonoTextStyleBuilder::new()
            .font(&FONT_10X20)
            .background_color(Rgb565::CSS_LIME)
            .text_color(Rgb565::BLACK)
            .build();

        Text::with_baseline("Call:", Point::new(LABEL_X, 40), label_style, Baseline::Top)
            .draw(dt)?;
        Text::with_baseline("SSID:", Point::new(LABEL_X, 80), label_style, Baseline::Top)
            .draw(dt)?;
        Text::with_baseline(
            "Icon:",
            Point::new(LABEL_X, 120),
            label_style,
            Baseline::Top,
        )
        .draw(dt)?;
        Text::with_baseline(
            "Freq (Hz):",
            Point::new(LABEL_X, 160),
            label_style,
            Baseline::Top,
        )
        .draw(dt)?;

        Text::with_baseline(
            "Brightness:",
            Point::new(LABEL_X, 200),
            label_style,
            Baseline::Top,
        )
        .draw(dt)?;

        self.callsign.selected = self.selected == 0;
        self.callsign.render(dt)?;

        self.ssid.valid = self.ssid().is_some();
        self.ssid.selected = self.selected == 1;
        self.ssid.render(dt)?;

        self.icon.valid = self.icon().is_some();
        self.icon.selected = self.selected == 2;
        self.icon.render(dt)?;

        self.freq.valid = self.freq().is_some();
        self.freq.selected = self.selected == 3;
        self.freq.render(dt)?;

        self.brightness.selected = self.selected == 4;
        self.brightness.render(dt)?;

        Ok(())
    }

    pub fn key_push(&mut self, k: crate::keyboard::KeyCode) {
        match self.selected {
            0 => self.callsign.key_push(k),
            1 => self.ssid.key_push(k),
            2 => self.icon.key_push(k),
            3 => self.freq.key_push(k),
            _ => unreachable!(),
        }
    }

    // updating brightness here
    // probably not perfect but whatever. At least we only update it when we scroll
    pub fn touchpad_scroll(&mut self, x: i8, y: i8, brightness: &mut u8) {
        match self.selected {
            0 => self.callsign.touchpad_scroll(x, 0),
            1 => self.ssid.touchpad_scroll(x, 0),
            2 => self.icon.touchpad_scroll(x, 0),
            3 => self.freq.touchpad_scroll(x, 0),
            4 => self.brightness.touchpad_scroll(x, 0),
            _ => unreachable!(),
        }

        let (_, scroll_y) = self.tracker.scroll(x, y);
        match scroll_y {
            ScrollAction::Previous => {
                if self.selected > 0 {
                    self.selected -= 1;
                }
            }
            ScrollAction::Next => {
                if self.selected < 4 {
                    self.selected += 1;
                }
            }
            ScrollAction::None => {}
        }

        *brightness = self.brightness.value()
    }
}
