use core::fmt::Write;

use embedded_graphics::{
    geometry::{Point, Size},
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::{Rgb565, RgbColor, WebColors},
    primitives::{Primitive, PrimitiveStyleBuilder, Rectangle},
    text::{Baseline, Text},
    Drawable,
};
use heapless::String;

use super::{
    scroll_tracker::{ScrollAction, ScrollTracker},
    textbox::{BORDER_WIDTH, CHAR_HEIGHT, CHAR_WIDTH},
    Element,
};

const SELECTED_COLOR: Rgb565 = Rgb565::WHITE;
const UNSELECTED_COLOR: Rgb565 = Rgb565::CSS_GRAY;

pub struct Slider {
    min: u8,
    max: u8,
    cur: u8,
    x: i32,
    y: i32,
    tracker: ScrollTracker,
    pub selected: bool,
}

impl Slider {
    pub fn new(min: u8, max: u8, x: i32, y: i32) -> Self {
        Self {
            min,
            max,
            cur: max,
            x,
            y,
            tracker: ScrollTracker::new_with_threshold(10),
            selected: true,
        }
    }

    pub fn value(&self) -> u8 {
        self.cur
    }

    fn width(&self) -> u32 {
        CHAR_WIDTH * 3
    }

    fn height(&self) -> u32 {
        CHAR_HEIGHT
    }
}

impl Element for Slider {
    type KeyPushReturn = ();

    fn render<
        E,
        DT: embedded_graphics::prelude::DrawTarget<
            Color = embedded_graphics::pixelcolor::Rgb565,
            Error = E,
        >,
    >(
        &mut self,
        dt: &mut DT,
    ) -> Result<(), E> {
        let bg_color = if self.selected {
            SELECTED_COLOR
        } else {
            UNSELECTED_COLOR
        };

        let rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(bg_color)
            .stroke_width(BORDER_WIDTH)
            .fill_color(bg_color)
            .build();
        Rectangle::new(
            Point::new(self.x, self.y),
            Size::new(self.width(), self.height()),
        )
        .into_styled(rect_style)
        .draw(dt)?;

        let text_style = MonoTextStyleBuilder::new()
            .font(&FONT_10X20)
            .text_color(Rgb565::BLACK)
            .build();

        let mut s: String<5> = String::new();
        write!(&mut s, "{}", self.cur).unwrap();
        Text::with_baseline(&s, Point::new(self.x, self.y), text_style, Baseline::Top).draw(dt)?;

        Ok(())
    }

    fn key_push(&mut self, _k: crate::keyboard::KeyCode) {}

    fn touchpad_scroll(&mut self, x: i8, y: i8) {
        let (scroll_x, _) = self.tracker.scroll(x, y);
        match scroll_x {
            ScrollAction::Next => {
                if self.cur > self.min {
                    self.cur -= 1;
                }
            }
            ScrollAction::Previous => {
                if self.cur < self.max {
                    self.cur += 1;
                }
            }
            ScrollAction::None => {}
        }
    }
}
