use core::{mem, ptr};

use embedded_graphics::{
    draw_target::DrawTarget,
    geometry::{Point, Size},
    mono_font::{ascii::FONT_10X20, MonoTextStyleBuilder},
    pixelcolor::{Rgb565, RgbColor},
    primitives::{Primitive, PrimitiveStyleBuilder, Rectangle},
    text::{Baseline, Text},
    Drawable,
};
use heapless::String;

use crate::keyboard::KeyCode;

use super::{
    scroll_tracker::{ScrollAction, ScrollTracker},
    Element,
};

const BORDER_COLOR: Rgb565 = Rgb565::new(28, 55, 28);
const INVALID_BORDER_COLOR: Rgb565 = Rgb565::new(28, 14, 7);
const BACKGROUND_COLOR: Rgb565 = Rgb565::new(0xFF, 0xFF, 0xFF);
pub const BORDER_WIDTH: u32 = 3;
pub const CHAR_WIDTH: u32 = 10;
pub const CHAR_HEIGHT: u32 = 20;

pub struct TextBox<const N: usize> {
    // in # of chars
    max_height: u32,
    char_width: u32,
    char_height: u32,
    cursor_x: u32,
    cursor_y: u32,
    cursor_i: usize,
    cur_line_len: u32,
    num_lines: u32,

    // in pixels
    x: i32,
    y: i32,

    number_only: bool,

    text: String<N>,
    pub selected: bool,
    pub valid: bool,

    tracker: ScrollTracker,
}

impl<const N: usize> TextBox<N> {
    pub fn new(
        x: i32,
        y: i32,
        char_width: u32,
        char_height: u32,
        max_height: u32,
        number_only: bool,
    ) -> Self {
        Self {
            max_height,
            char_width,
            char_height,
            cursor_x: 0,
            cursor_y: 0,
            cursor_i: 0,
            cur_line_len: 0,
            num_lines: 1,
            number_only,

            x,
            y,

            text: String::new(),
            selected: true,
            valid: true,

            tracker: ScrollTracker::new(),
        }
    }

    pub fn set_text(&mut self, text: &str) -> Result<(), ()> {
        self.text = text.try_into()?;

        for c in text.chars() {
            match c {
                '\n' => {
                    self.cursor_x = 0;
                    self.cursor_y += 1;
                    self.num_lines += 1;
                    self.cur_line_len = 0;
                }
                _ => {
                    self.cursor_inc_char();
                }
            }

            self.cursor_i += 1;
        }

        Ok(())
    }

    pub fn get_text(&self) -> &str {
        &self.text
    }

    pub fn clear(&mut self) -> String<N> {
        let mut other = String::new();
        mem::swap(&mut self.text, &mut other);

        self.cursor_x = 0;
        self.cursor_y = 0;
        self.cursor_i = 0;
        self.num_lines = 1;
        self.cur_line_len = 0;

        other
    }

    fn width(&self) -> u32 {
        BORDER_WIDTH * 2 + CHAR_WIDTH * self.char_width
    }

    fn char_height(&self) -> u32 {
        self.char_height.max(self.num_lines).min(self.max_height)
    }

    pub fn height(&self) -> u32 {
        BORDER_WIDTH * 2 + CHAR_HEIGHT * self.char_height()
    }

    fn y(&self) -> i32 {
        self.y - i32::try_from((self.char_height() - 1) * CHAR_HEIGHT).unwrap()
    }

    fn cursor_inc_char(&mut self) {
        self.cursor_x += 1;
        self.cur_line_len += 1;
        if self.cursor_x >= self.char_width {
            self.cursor_x = 0;
            self.cursor_y += 1;
            self.cur_line_len = 0;
            self.num_lines += 1;
        }
    }

    fn cursor_dec_char(&mut self) {
        if self.cursor_x > 0 {
            self.cursor_x -= 1;
        } else {
            self.cursor_y -= 1;
            self.num_lines -= 1;
            self.cursor_x = self.calc_cursor_pos_after_backspace();
        }
    }

    // returns x pos
    fn calc_cursor_pos_after_backspace(&self) -> u32 {
        let mut i = self.cursor_i - 1;
        let mut count = 0;
        while i > 0 {
            i -= 1;
            count += 1;
            // Inefficient; may want to rewrite this if text is laggy
            // O(n)
            if self.text.chars().nth(i).unwrap() == '\n' {
                count -= 1;
                break;
            }
        }

        count % self.char_width
    }

    fn insert_char(&mut self, c: char) -> bool {
        insert_char_at_pos(&mut self.text, self.cursor_i, c)
    }
}

impl<const N: usize> Element for TextBox<N> {
    type KeyPushReturn = ();

    fn render<E, DT: DrawTarget<Color = Rgb565, Error = E>>(
        &mut self,
        dt: &mut DT,
    ) -> Result<(), E> {
        // Draw border/background
        let border_color = if self.valid {
            BORDER_COLOR
        } else {
            INVALID_BORDER_COLOR
        };
        let rect_style = PrimitiveStyleBuilder::new()
            .stroke_color(border_color)
            .stroke_width(BORDER_WIDTH)
            .fill_color(BACKGROUND_COLOR)
            .build();
        Rectangle::new(
            Point::new(self.x, self.y()),
            Size::new(self.width(), self.height()),
        )
        .into_styled(rect_style)
        .draw(dt)?;

        // Draw text
        let text_x = self.x + i32::try_from(BORDER_WIDTH).unwrap();
        let text_y = self.y() + i32::try_from(BORDER_WIDTH).unwrap();

        let text_style = MonoTextStyleBuilder::new()
            .font(&FONT_10X20)
            .text_color(Rgb565::BLACK)
            .build();

        let mut cur_x = 0;
        let mut cur_y = 0;
        let mut buf = [0; 4];
        for c in self.text.chars() {
            if c == '\n' {
                cur_x = 0;
                cur_y += 1;

                continue;
            }

            Text::with_baseline(
                c.encode_utf8(&mut buf),
                Point::new(
                    text_x + i32::try_from(cur_x * CHAR_WIDTH).unwrap(),
                    text_y + i32::try_from(cur_y * CHAR_HEIGHT).unwrap(),
                ),
                text_style,
                Baseline::Top,
            )
            .draw(dt)?;

            cur_x += 1;
            if cur_x == self.char_width {
                cur_x = 0;
                cur_y += 1;
            }
        }

        // Draw cursor
        if self.selected {
            Text::with_baseline(
                "_",
                Point::new(
                    text_x + i32::try_from(self.cursor_x * CHAR_WIDTH).unwrap(),
                    text_y + i32::try_from(self.cursor_y * CHAR_HEIGHT).unwrap(),
                ),
                text_style,
                Baseline::Top,
            )
            .draw(dt)?;
        }

        Ok(())
    }

    fn key_push(&mut self, k: KeyCode) {
        match k {
            KeyCode::Backspace => {
                if self.cursor_i == 0 {
                    return;
                }

                if let Some(popped) = remove_char_at_pos(&mut self.text, self.cursor_i - 1) {
                    if popped == '\n' {
                        self.cursor_x = self.calc_cursor_pos_after_backspace();
                        self.cur_line_len = self.cursor_x;
                        self.cursor_y -= 1;
                        self.num_lines -= 1;
                    } else {
                        self.cursor_dec_char();
                        self.cur_line_len -= 1;
                    }
                    self.cursor_i -= 1;
                }
            }
            KeyCode::Enter => {
                let _ = self.text.push('\n');
                self.cursor_x = 0;
                self.cursor_y += 1;
                self.num_lines += 1;
                self.cursor_i += 1;
                self.cur_line_len = 0;
            }
            c => {
                let c = if self.number_only {
                    c.as_number()
                } else {
                    c.as_char()
                };

                if let Some(c) = c {
                    if self.insert_char(c) {
                        self.cursor_inc_char();
                        self.cursor_i += 1;
                    }
                }
            }
        }
    }

    fn touchpad_scroll(&mut self, x: i8, y: i8) {
        let (scroll_x, _) = self.tracker.scroll(x, y);
        match scroll_x {
            ScrollAction::Next => {
                if self.cursor_x > 0 {
                    self.cursor_x -= 1;
                    self.cursor_i -= 1;
                }
            }
            ScrollAction::Previous => {
                if self.cursor_x < self.cur_line_len {
                    self.cursor_x += 1;
                    self.cursor_i += 1;
                }
            }
            ScrollAction::None => {}
        }
    }
}

fn insert_char_at_pos<const N: usize>(s: &mut String<N>, i: usize, c: char) -> bool {
    let idx = if i == 0 {
        0
    } else {
        let (idx, c) = s.char_indices().nth(i - 1).unwrap();

        idx + c.len_utf8()
    };

    assert!(s.is_char_boundary(idx));

    let mut bytes = [0; 4];
    let bytes = c.encode_utf8(&mut bytes);

    let len = s.len();
    let amt = bytes.len();
    if s.capacity() < len + amt {
        return false;
    }

    unsafe {
        let v = s.as_mut_vec();
        ptr::copy(
            v.as_ptr().add(idx),
            v.as_mut_ptr().add(idx + amt),
            len - idx,
        );
        ptr::copy_nonoverlapping(bytes.as_ptr(), v.as_mut_ptr().add(idx), amt);
        v.set_len(len + amt);
    }

    true
}

fn remove_char_at_pos<const N: usize>(s: &mut String<N>, i: usize) -> Option<char> {
    let (idx, c) = s.char_indices().nth(i)?;
    let len = s.len();
    let amt = c.len_utf8();

    unsafe {
        let v = s.as_mut_vec();
        ptr::copy(
            v.as_ptr().add(idx + amt),
            v.as_mut_ptr().add(idx),
            len - idx - amt,
        );
        v.set_len(len - amt);
    }

    Some(c)
}
