use heapless::{FnvIndexSet, String, Vec};

use super::{
    contact::CONTACT_DATA_END, flash::FlashManager, Storage, MAX_CONTACT_CALLSIGN_LENGTH, X25,
};

const MESSAGE_HEADER_START: usize = CONTACT_DATA_END;
// Includes 2 byte checksum
const MESSAGE_HEADER_SIZE: usize = 6;
pub const MAX_MESSAGE_BODY_LENGTH: usize = 500;
// 1 byte callsign length
// 2 byte message length
// 1 byte SSID
// 1 byte direction + received
// 1 byte send attempts
// 1 byte ack number
// 2 byte checksum
const MESSAGE_LENGTH: usize =
    1 + 2 + 1 + 1 + 1 + 1 + MAX_MESSAGE_BODY_LENGTH + MAX_CONTACT_CALLSIGN_LENGTH + 2;
const MAX_MESSAGES: usize = 1024;
pub const MESSAGE_HEADER_END: usize =
    MESSAGE_HEADER_START + MESSAGE_HEADER_SIZE + MAX_MESSAGES * MESSAGE_LENGTH;

const MAX_RETRIES: u8 = 5;

#[derive(Copy, Clone)]
pub struct MessageGroup {
    header: MessageHeader,
}

impl MessageGroup {
    pub(crate) fn new(manager: &FlashManager) -> Self {
        Self {
            header: MessageHeader::read(manager),
        }
    }

    pub fn len(&self) -> u16 {
        self.header.len
    }

    pub fn push_message(&mut self, manager: &mut FlashManager, message: Message) -> u16 {
        let id;

        if usize::from(self.header.len) == MAX_MESSAGES {
            // overwrite the first message
            id = self.header.start.into();
            message.write(manager, id);
            self.header.start += 1;
            self.header.start %= self.header.len;
        } else {
            // append to end
            id = self.header.len.into();
            message.write(manager, id);
            self.header.len += 1;
        };

        self.header.write(manager);

        id.try_into().unwrap()
    }

    pub fn get_message(&self, manager: &FlashManager, id: u16) -> Option<Message> {
        // This only works because we never remove from queue
        assert!(id < self.header.len);

        Message::read(manager, id.into())
    }

    pub fn iter<'a, 'b, 'c>(
        &'a self,
        storage: &'b Storage,
        callsign: &'c str,
        ssid: u8,
    ) -> MessageIter<'a, 'b, 'c> {
        MessageIter::new(self, storage, callsign, ssid)
    }

    pub fn iter_addressees<'a, 'b>(&'a self, manager: &'b FlashManager) -> AddresseeIter<'a, 'b> {
        AddresseeIter::new(self, manager)
    }
}

#[derive(Copy, Clone)]
pub struct MessageHeader {
    start: u16,
    len: u16,
}

impl MessageHeader {
    fn read(manager: &FlashManager) -> Self {
        let bytes: [_; MESSAGE_HEADER_SIZE] = manager.read_slice(MESSAGE_HEADER_START);
        let start = u16::from_le_bytes([bytes[0], bytes[1]]);
        let len = u16::from_le_bytes([bytes[2], bytes[3]]);
        let checksum_expected = u16::from_le_bytes([bytes[4], bytes[5]]);
        let checksum_actual = X25.checksum(&bytes[0..4]);

        if checksum_expected == checksum_actual {
            Self { start, len }
        } else {
            Self { start: 0, len: 0 }
        }
    }

    fn write(&self, manager: &mut FlashManager) {
        let mut buf = [0; MESSAGE_HEADER_SIZE];
        buf[0..2].copy_from_slice(&self.start.to_le_bytes());
        buf[2..4].copy_from_slice(&self.len.to_le_bytes());
        let checksum = X25.checksum(&buf[0..4]);
        buf[4..].copy_from_slice(&checksum.to_le_bytes());

        manager.write_data(MESSAGE_HEADER_START, &buf);
    }
}

pub struct Message {
    // Storing the entire callsign/ssid per message is kind of a waste
    // but it's simple. We can fix it later if it becomes an issue
    callsign: String<MAX_CONTACT_CALLSIGN_LENGTH>,
    ssid: u8,
    message: String<MAX_MESSAGE_BODY_LENGTH>,
    direction: MessageDirection,
    pub received: bool,
    pub send_attempts: u8,
    // The ack number we received with this. Used so we can detect duplicates
    pub ack_number: u8,
}

impl Message {
    pub fn new(
        callsign: String<MAX_CONTACT_CALLSIGN_LENGTH>,
        ssid: u8,
        message: String<MAX_MESSAGE_BODY_LENGTH>,
        direction: MessageDirection,
    ) -> Self {
        Self {
            callsign,
            ssid,
            message,
            direction,
            received: false,
            send_attempts: 0,
            ack_number: 0,
        }
    }

    pub fn callsign(&self) -> &str {
        &self.callsign
    }

    pub fn ssid(&self) -> u8 {
        self.ssid
    }

    pub fn content(&self) -> &str {
        &self.message
    }

    pub fn direction(&self) -> MessageDirection {
        self.direction
    }

    pub fn failed(&self) -> bool {
        self.send_attempts >= MAX_RETRIES
    }

    fn read(manager: &FlashManager, idx: usize) -> Option<Self> {
        let start_addr = MESSAGE_HEADER_START + MESSAGE_HEADER_SIZE + idx * MESSAGE_LENGTH;

        let bytes = manager.read_slice::<MESSAGE_LENGTH>(start_addr);
        let checksum_expected =
            u16::from_le_bytes([bytes[MESSAGE_LENGTH - 2], bytes[MESSAGE_LENGTH - 1]]);
        let checksum_actual = X25.checksum(&bytes[..(MESSAGE_LENGTH - 2)]);

        if checksum_expected != checksum_actual {
            return None;
        }

        let callsign_length: usize = bytes[0].into();
        let message_length: usize = u16::from_le_bytes([bytes[1], bytes[2]]).into();
        let ssid = bytes[3];
        let direction = MessageDirection::from(bytes[4] & 0x01 > 0);
        let received = bytes[4] & 0x02 > 0;
        let send_attempts = bytes[5];
        let ack_number = bytes[6];

        let remainder = &bytes[7..];
        let callsign_bytes = &remainder[..callsign_length];
        let message_bytes = &remainder[MAX_CONTACT_CALLSIGN_LENGTH..][..message_length];

        let callsign = String::from_utf8(Vec::from_slice(callsign_bytes).ok()?).ok()?;
        let message = String::from_utf8(Vec::from_slice(message_bytes).ok()?).ok()?;

        Some(Self {
            callsign,
            ssid,
            message,
            direction,
            received,
            send_attempts,
            ack_number,
        })
    }

    pub fn write(&self, manager: &mut FlashManager, idx: usize) {
        let start_addr = MESSAGE_HEADER_START + MESSAGE_HEADER_SIZE + idx * MESSAGE_LENGTH;

        let callsign_bytes = self.callsign.as_bytes();
        let message_bytes = self.message.as_bytes();

        let mut buf = [0; MESSAGE_LENGTH];
        buf[0] = callsign_bytes.len().try_into().unwrap();
        buf[1..3].copy_from_slice(&u16::try_from(self.message.len()).unwrap().to_le_bytes());
        buf[3] = self.ssid;
        buf[4] = (u8::from(self.received) << 1) | u8::from(bool::from(self.direction));
        buf[5] = self.send_attempts;
        buf[6] = self.ack_number;

        let remainder = &mut buf[7..];
        remainder[..callsign_bytes.len()].copy_from_slice(callsign_bytes);
        remainder[MAX_CONTACT_CALLSIGN_LENGTH..][..(message_bytes.len())]
            .copy_from_slice(message_bytes);

        let checksum = X25.checksum(&buf[..(MESSAGE_LENGTH - 2)]);
        buf[(MESSAGE_LENGTH - 2)..].copy_from_slice(&checksum.to_le_bytes());

        manager.write_data(start_addr, &buf);
    }
}

#[derive(Copy, Clone, PartialEq, Eq)]
pub enum MessageDirection {
    From,
    To,
}

impl From<bool> for MessageDirection {
    fn from(value: bool) -> Self {
        if value {
            MessageDirection::From
        } else {
            MessageDirection::To
        }
    }
}

impl From<MessageDirection> for bool {
    fn from(value: MessageDirection) -> Self {
        match value {
            MessageDirection::From => true,
            MessageDirection::To => false,
        }
    }
}

#[derive(Clone)]
pub struct MessageIter<'a, 'b, 'c> {
    group: &'a MessageGroup,
    manager: &'b FlashManager,
    callsign: &'c str,
    ssid: u8,
    i: u16,
    j: u16,
}

impl<'a, 'b, 'c> MessageIter<'a, 'b, 'c> {
    pub fn new(group: &'a MessageGroup, storage: &'b Storage, callsign: &'c str, ssid: u8) -> Self {
        Self {
            group,
            manager: storage.flash(),
            callsign,
            ssid,
            i: 0,
            j: storage.messages_len(),
        }
    }
}

impl<'a, 'b, 'c> Iterator for MessageIter<'a, 'b, 'c> {
    type Item = Message;

    fn next(&mut self) -> Option<Self::Item> {
        while self.i < self.j {
            let idx = (self.i + self.group.header.start) % self.group.header.len;
            self.i += 1;

            if let Some(msg) = Message::read(self.manager, idx.into()) {
                if msg.callsign == self.callsign && msg.ssid == self.ssid {
                    return Some(msg);
                }
            }
        }

        None
    }
}

impl<'a, 'b, 'c> DoubleEndedIterator for MessageIter<'a, 'b, 'c> {
    fn next_back(&mut self) -> Option<Self::Item> {
        while self.i < self.j {
            let idx = (self.j - 1 + self.group.header.start) % self.group.header.len;
            self.j -= 1;

            if let Some(msg) = Message::read(self.manager, idx.into()) {
                if msg.callsign == self.callsign && msg.ssid == self.ssid {
                    return Some(msg);
                }
            }
        }

        None
    }
}

// Orders from newest to oldest
pub struct AddresseeIter<'a, 'b> {
    group: &'a MessageGroup,
    manager: &'b FlashManager,
    i: u16,
    seen: FnvIndexSet<(String<MAX_CONTACT_CALLSIGN_LENGTH>, u8), MAX_MESSAGES>,
}

impl<'a, 'b> AddresseeIter<'a, 'b> {
    pub fn new(group: &'a MessageGroup, manager: &'b FlashManager) -> Self {
        Self {
            group,
            manager,
            i: group.header.len,
            seen: FnvIndexSet::new(),
        }
    }
}

impl<'a, 'b> Iterator for AddresseeIter<'a, 'b> {
    type Item = (String<MAX_CONTACT_CALLSIGN_LENGTH>, u8);

    fn next(&mut self) -> Option<Self::Item> {
        while self.i > 0 {
            self.i -= 1;

            let idx = (self.i + self.group.header.start) % self.group.header.len;
            if let Some(msg) = Message::read(self.manager, idx.into()) {
                let ident = (msg.callsign, msg.ssid);
                if !self.seen.contains(&ident) {
                    self.seen.insert(ident.clone()).unwrap();

                    return Some(ident);
                }
            }
        }

        None
    }
}
