mod contact;
mod flash;
mod message;
mod settings;

use contact::ContactGroup;
pub use contact::{
    Contact, ContactIter, MAX_CONTACTS, MAX_CONTACT_CALLSIGN_LENGTH, MAX_CONTACT_NAME_LENGTH,
};
use flash::FlashManager;
pub use message::{Message, MessageDirection, MessageGroup, MessageIter, MAX_MESSAGE_BODY_LENGTH};
pub use settings::Settings;

use self::message::AddresseeIter;

const X25: crc::Crc<u16> = crc::Crc::<u16>::new(&crc::CRC_16_IBM_SDLC);

pub struct Storage {
    manager: FlashManager,
    contacts: ContactGroup,
    messages: MessageGroup,
    settings: Settings,
}

impl Storage {
    pub(crate) fn new() -> Self {
        let manager = FlashManager::new();
        let contacts = ContactGroup::new(&manager);
        let messages = MessageGroup::new(&manager);
        let settings = Settings::new(&manager);

        Self {
            manager,
            contacts,
            messages,
            settings,
        }
    }

    pub(crate) fn flash(&self) -> &FlashManager {
        &self.manager
    }

    pub(crate) fn flash_mut(&mut self) -> &mut FlashManager {
        &mut self.manager
    }

    // Punch through our contact methods

    pub fn new_contact(&self) -> Contact {
        self.contacts.new_contact()
    }

    pub fn save_contact(&mut self, contact: Contact) {
        self.contacts.save_contact(&mut self.manager, contact);
    }

    pub fn delete_contact(&mut self, contact_id: usize) {
        self.contacts.delete_contact(&mut self.manager, contact_id);
    }

    pub fn contacts(&self) -> ContactIter {
        self.contacts.contacts(&self.manager)
    }

    pub fn contacts_len(&self) -> usize {
        self.contacts.len()
    }

    // Punch through our message methods

    pub fn push_message(&mut self, message: Message) -> u16 {
        self.messages.push_message(&mut self.manager, message)
    }

    pub fn get_message(&self, id: u16) -> Option<Message> {
        self.messages.get_message(&self.manager, id)
    }

    pub fn messages<'a, 'c>(&'a self, callsign: &'c str, ssid: u8) -> MessageIter<'a, 'a, 'c> {
        self.messages.iter(self, callsign, ssid)
    }

    pub fn messages_len(&self) -> u16 {
        self.messages.len()
    }

    pub fn message_addressees(&self) -> AddresseeIter {
        self.messages.iter_addressees(&self.manager)
    }

    // Settings

    pub fn settings(&self) -> &Settings {
        &self.settings
    }

    pub fn settings_mut(&mut self) -> &mut Settings {
        &mut self.settings
    }

    pub fn save_settings(&mut self) {
        self.settings.write(&mut self.manager);
    }
}
