use crate::state::InvalidState;
use core::fmt::Debug;

pub enum BusError<SE> {
    Timeout,
    Spi(SE),
}

impl<SE: Debug> Debug for BusError<SE> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            BusError::Timeout => write!(f, "BusError::Timeout"),
            BusError::Spi(x) => write!(f, "BusError::Spi({x:?})"),
        }
    }
}

pub enum RfError<SE> {
    InvalidState(InvalidState),
    Bus(BusError<SE>),
}

impl<SE: Debug> From<InvalidState> for RfError<SE> {
    fn from(is: InvalidState) -> Self {
        Self::InvalidState(is)
    }
}

pub enum TransferError<SE> {
    // also handles underflows
    FifoOverflow,
    // Expected a packet of length N, however
    // the chip gave us >N bytes
    TooMuchData,
    Bus(BusError<SE>),
}

impl<SE: Debug> Debug for TransferError<SE> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            TransferError::FifoOverflow => write!(f, "TransferError::FifoOverflow"),
            TransferError::TooMuchData => write!(f, "TransferError::TooMuchData"),
            TransferError::Bus(x) => write!(f, "TransferError::Bus({x:?})"),
        }
    }
}

pub enum TxError<SE> {
    // Given slice was bigger than the maximum packet size
    TooMuchData,
    Bus(BusError<SE>),
}

impl<SE: Debug> Debug for TxError<SE> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            TxError::TooMuchData => write!(f, "TransferError::TooMuchData"),
            TxError::Bus(x) => write!(f, "TransferError::Bus({x:?})"),
        }
    }
}

pub(crate) trait SpiErrorToOtherError<T, E> {
    fn be(self) -> Result<T, BusError<E>>;
}

impl<T, E> SpiErrorToOtherError<T, E> for Result<T, E> {
    fn be(self) -> Result<T, BusError<E>> {
        self.map_err(|e| BusError::Spi(e))
    }
}

pub(crate) trait BusErrorToOtherError<T, E> {
    fn re(self) -> Result<T, RfError<E>>;

    fn te(self) -> Result<T, TransferError<E>>;

    fn txe(self) -> Result<T, TxError<E>>;
}

impl<T, E> BusErrorToOtherError<T, E> for Result<T, BusError<E>> {
    fn re(self) -> Result<T, RfError<E>> {
        self.map_err(|e| RfError::Bus(e))
    }

    fn te(self) -> Result<T, TransferError<E>> {
        self.map_err(|e| TransferError::Bus(e))
    }

    fn txe(self) -> Result<T, TxError<E>> {
        self.map_err(|e| TxError::Bus(e))
    }
}
